import pickle as cPickle
import time
import torch as t
import numpy as np
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader

def predictCollate(batch):
	x1 = []
	x2 = []		
	sx1 = []
	sx2 = []
	for i in  batch:		
		x1.append(i[0])
		x2.append(i[1])
		if len(i) > 2:
			sx1.append(i[2])
			sx2.append(i[3])
	if len(sx1) > 0 or len(sx2) > 0:	
		#print sx1[0]
		return t.LongTensor(x1), t.LongTensor(x2), t.FloatTensor(np.vstack(sx1)), t.FloatTensor(np.vstack(sx2))
	else:
		return t.LongTensor(x1), t.LongTensor(x2)

def predictMetaCollate(batch):
	x1 = []
	x2 = []
	side1 = []
	side2 = []
	for i in  batch:
		#assert i[0].shape == i[1].shape == i[2].shape
		x1.append(i[0])
		x2.append(i[1])
		if len(batch[0]) > 3:
			if len(i[2]) > 0:
				side1.append(i[2])
			if len(i[3]) > 0:
				side2.append(i[3])
	#print x1
	#print y
	x1 = np.hstack(x1)
	x2 = np.hstack(x2)
	#print len(side1), type(side1)
	if len(batch[0]) > 3:
		if len(side1) == 0:
			side1 = None
		else:
			side1 = np.vstack(side1)
			#print side1.size()
			#assert side1.size()[0] == len(x1)
		if len(side2) == 0:
			side2 = None
		else:
			tmp1 = np.vstack(side2)
			#print tmp1.shape
			side2 = tmp1
			#print "qui",side2.size()
			#assert side2.size()[0] == len(x2)
		return x1, x2, side1, side2
	else:
		return x1, x2

def metaCollate(batch):
	x1 = []
	x2 = []
	y = []
	side1 = []
	side2 = []
	for i in  batch:
		#assert i[0].shape == i[1].shape == i[2].shape
		x1.append(i[0])
		x2.append(i[1])
		y.append(i[2])
		if len(batch[0]) > 3:
			if len(i[3]) > 0:
				side1.append(i[3])
			if len(i[4]) > 0:
				side2.append(i[4])
	#print x1
	#print y
	x1 = np.hstack(x1)
	x2 = np.hstack(x2)
	y = np.vstack(y)
	#print len(side1), type(side1)
	if len(batch[0]) > 3:
		if len(side1) == 0:
			side1 = None
		else:
			side1 = np.vstack(side1)
			#print side1.size()
			#assert side1.size()[0] == len(x1)
		if len(side2) == 0:
			side2 = None
		else:
			side2 = np.vstack(side2)
			#print tmp1.shape
			#side2 = t.FloatTensor(tmp1).squeeze()
			#print "qui",side2.size()
			#assert side2.size()[0] == len(x2)
		return x1, x2, y, side1, side2
	else:
		return x1, x2, y




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pytorchLossUtils.py
#  
#  Copyright 2018 Daniele Raimondi <daniele.raimondi@vub.be>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import numpy as np
import torch as t
from torch.autograd import Variable
from collections import Iterable
from multipledispatch import dispatch

def SafeVariable(v):
	""" 
		Utility function that helps checking types.

		Parameters
		----------
		v : object
			Input variable

		Returns
		-------
		Returns None if v is none, raises exception if v is 0 and returns v otherwise.
	:meta private:
	"""

	if type(v) == type(None):
		return None
	elif len(v) == 0:
		raise Exception("Very strange to have empty vars here")
	else:
		return v

def f():
	"""
	:meta private:
	"""
	pass

class Entity(object):

	"""
	Class representing the Entity concept. 
	"""
	#@dispatch(str, list, type)
	def __init__(self, name, domain, dtype=np.int32):
		""" 
			Constructor for the Entity class.

			Parameters
			----------
			name : str
				Name of the Entity (use a mnemonic name describing the class of objects represented by the Entity)
			domain : iterable (list) containing str
				List of the possible objects belonging to this class (e.g. patients IDs, proteins Uniprot identifiers, ...). It is an unique identifier naming (with a string) all the objects composing the domain of the Entity.
			dtype : np.dtype
				Smallest possible numpy type able to uniquely enumerate all the objects. len(domain) < max_number_representable(dtype).
			
			Returns
			-------
		"""
		assert isinstance(domain, Iterable)
		assert len(domain) > 0
		self.dtype = dtype
		self.name = name
		self.domain = domain
		self.lookup = {}
		i = 0
		while i < len(domain):
			self.lookup[domain[i]] = i
			i+=1
		assert len(self.lookup) == len(self.domain)

	def __next__(self):
		return self.domain.next()

	next = __next__

	def __iter__(self):
		return iter(self.domain)

	@dispatch(str)
	def __getitem__(self, x):
		""" Method that	returns the numeric value internally associated to each object in the Entity class.

			Parameters
			----------
			x : str
				String name of a specific object in the Entity.

			Returns
			primary key : int
			-------
		"""
		return self.lookup[x]
	
	@dispatch(int)
	def __getitem__(self, x):
		""" Method that	returns the str name of the object with primary key x.

			Parameters
			----------
			x : int
				Primary key (unique id) of an object in the domain represented by the Entity.

			Returns
			name of the object : str
			-------
		"""
		#print(x, len(self.domain))
		assert x < len(self.domain)
		return x

	@dispatch(np.int16)
	def __getitem__(self, x):
		""" Method that	returns the str name of the object with primary key x.

			Parameters
			----------
			x : int
				Primary key (unique id) of an object in the domain represented by the Entity.

			Returns
			name of the object : str
			-------
		"""
		#print(x, len(self.domain))
		assert x < len(self.domain)
		return x

	@dispatch(np.int64)
	def __getitem__(self, x):
		""" Method that	returns the str name of the object with primary key x.

			Parameters
			----------
			x : int
				Primary key (unique id) of an object in the domain represented by the Entity.

			Returns
			name of the object : str
			-------
		"""
		#print(x, len(self.domain))
		assert x < len(self.domain)
		return x

	@dispatch(np.int32)
	def __getitem__(self, x):
		""" Method that	returns the str name of the object with primary key x.

			Parameters
			----------
			x : int
				Primary key (unique id) of an object in the domain represented by the Entity.

			Returns
			name of the object : str
			-------
		"""
		#print(x, len(self.domain))
		assert x < len(self.domain)
		return x
		#return self.domain[x]
	
	def has_key(self, x):
		return x in self.lookup

	def __len__(self):
		return len(self.domain)

class Relation(dict):
	"""
	Class that represent a relation (matrix in MF terms) with all its parameters and functions.

	
	"""
	#@dispatch(str, Entity, Entity, DM.DataMatrix, str, )
	def __init__(self, name:str, domain1:Entity, domain2:Entity, data, task:str, loss:t.nn.Module, relationWeight:float, side1=None, side2 = None, path=None):
		""" Constructor for the Relation class..

			Parameters
			----------
			name : str
				Mnemonic name of the specific relation/matrix.
			domain1 : Entity
				Entity1 involved in the relation (on dimension 0)
			domain2 : Entity
				Entity2 involved in the relation (on dimension 1)
			data : DataMatrix
				DataMatrix object containing the matrix describing this relation
			task : str ["regression", "binary"]
				Type of prediction task associated to this relation. "Regression" for real valued predictions, "binary" for binary classification.
			loss : NX.NXLosses or t.nn.Module
				Pytorch-like loss module corresponding to the loss that must be used to compute the reconstruction error for this relation.
			relationWeight : float
				A relation-specific weight that will multiply the loss score during training.

			Returns
			-------
		"""
		super(Relation, self).__init__({})	
		assert type(name) == str
		self["name"] = name
		#assert type1 == "DS" or type1 == "OD"
		#d["type"] = type1
		#assert type(domain1) == Entity
		self["domain1"] = domain1
		assert type(domain2) == Entity
		self["domain2"] = domain2
		self["data"] = data
		assert task == "regression" or task == "binary"
		self["task"] = task
		assert relationWeight == "relativeToTarget" or type(relationWeight) == int or type(relationWeight) == float
		self["relationWeight"] = relationWeight
		self["loss"] = loss

class MetaRelation(object): #it's basically a list with constraints on what you can add

	"""Constructor for the Relation class. The Meta Relation represents multi-relations between the same entities (used for example in tensor factorization).

	As a convention, we recommend to use names in the form ENT1-ENT2.

	The domains must be the same for each relation in it, since the MetaRelation defines a tensor where the dimension 0 and 1 represent the same entities for all the matrices involved in the tensor.
	Can allow side info (common to all relations in it).

	"""

	def __init__(self, name,  domain1, domain2, side1=None, side2=None, relations=[], prediction=False):
		"""Constructor for the MetaRelation class.

		Parameters
		----------
		name : str
			Mnemonic name of the specific relation/matrix.
		domain1 : Entity
			Entity1 involved in the relation (on dimension 0)
		domain2 : Entity
			Entity2 involved in the relation (on dimension 1)
		relations : list of NX.Relation objects
			List of the relations involved in this MetaRelation (tensor)
		
		Returns
		-------
		"""
		self.name = name
		if side2 != None:
			assert len(side2) == len(domain2)
		self.side2 = side2
		if side1 != None:
			assert len(side1) == len(domain1)
		self.side1 = side1
		assert type(domain1) == Entity
		self.domain1 = domain1
		assert type(domain2) == Entity
		self.domain2 = domain2

		self.relationList = []
		for r in relations:
			self.append(r)

	def append(self, r):
		"""
		Method that adds a Relation object to an existing MetaRelation
		Parameters
		----------
		r : Relation
				
		Returns
		-------
		"""

		#print type(r)
		assert isinstance(r, Relation)
		assert self.domain1 == r["domain1"]
		assert self.domain2 == r["domain2"]
		self.relationList.append(r)

	#@dispatch(str)
	def getPos(self, x):
		for c, i in enumerate(self.relationList):
			if i["name"] == x:
				return c
		print ("MetaRelation %s does not contain relation named %s"% (self.name, x))
		return None


	@dispatch(str)
	def __getitem__(self, x:str)->Relation:
		"""
		Getitem method that searches by Relation.name

		Parameters
		----------
		x : str
			The name of a Relation in this MetaRelation

		Returns
		-------
		The target Relation or None
		"""

		for i in self.relationList:
			if i["name"] == x:
				return i
		print ("MetaRelation %s does not contain relation named %s"% (self.name, x))
		return None

	@dispatch(int)
	def __getitem__(self, x:int) -> Relation:
		"""
		Getitem method that searches by position of the target Relation in the tensor/MetaRelation.

		Parameters
		----------
		x : int
			The position (index) of a Relation in this MetaRelation

		Returns
		-------
		The target Relation or None
		"""

		return self.relationList[x]

	def __next__(self):
		return self.relationList.next()

	next = __next__

	def __iter__(self):
		return iter(self.relationList)

	def pop(self, pos):
		return self.relationList.pop(pos)

	def __len__(self):
		return len(self.relationList)

class ERgraph(list):
	"""
	Class that represents the entire Entity-Relation graphs, namely a list of MetaRelations. Each MetaRelation might contain multiple Relations.
	"""
	def __init__(self, entityList:list, name = ""):
		"""
		Constructor for the ERgraph (Entity-Relation Graph) object.

		Parameters
		----------
		entityList : list of MetaRelations
			List of MetaRelations that describe the full Entity Relation graph.

		Returns
		-------
		"""

		self.graph = []
		self.name = name
		self.lookup = {}
		for i, rel in enumerate(entityList):
			assert type(rel) == MetaRelation
			self.graph.append({"name":rel.name, "pos": i, "lenDomain1":len(rel.domain1), "lenDomain2":len(rel.domain2), "arity":len(rel), "metaRelation":rel}) #TODO:add side
			self.lookup[rel.name] = self.graph[i]
		super(ERgraph, self).__init__(entityList)	

	def __getitem__(self, x):
		if type(x) == int:
			return super(ERgraph, self).__getitem__(x)
		elif type(x) == str:
			return self.lookup[x]
		else:
			raise Exception("Unexpected:  ", x)

	@dispatch(MetaRelation)
	def __contains__(self, x):
		"""
		Function that determines whether a specific MetaRelation.name is present in the graph.

		Parameters
		----------
		x : MetaRelation 
			A MetaRelation object.

		Returns
		-------
		bool (Is x present?)
		"""

		#if type(x) == int:
		#	return super(ERgraph, self).__contains__(x)
		return x in self


	@dispatch(str)
	def __contains__(self, x):
		"""
		Function that determines whether a specific MetaRelation.name is present in the graph.

		Parameters
		----------
		x : str (MetaRelation name)
			Name of a MetaRelation object.

		Returns
		-------
		bool (Is x present?)
		"""

		#if type(x) == int:
		#	return super(ERgraph, self).__contains__(x)
		if type(x) == str:
			return x in self.lookup
		else:
			raise Exception("Unexpected: ", x)

	def __str__(self):
		"""
		Function that expresses the ERgraph as string 

		Parameters
		----------
		
		Returns
		-------
		"""

		s = "ERgraph:\n-Name:%s\n" % self.name
		for r in self.graph:
			s+= "\tMetaRel:%s, dom1:%d, dom2:%d, arity:%d\n" % (r["name"], r["lenDomain1"], r["lenDomain2"], r["arity"])
		return s





import pickle as cPickle
import numpy as np
from multipledispatch import dispatch
import time
import NXTfusion.NXTfusion as NX #import Entity
from scipy.sparse import coo_matrix

class SideInfo(object):

	"""
	Class that encapsulated the side information raw data in order to be efficiently processed by NXTfusion. You can use this class to wrap side information vectors analogously to how DataMatrix wraps matrix/relations.
	"""

	@dispatch(str,  NX.Entity, dict)
	def __init__(self, name, ent1, data):
		"""
			One of the alternative constructors for the SideInfo class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			data: dict
				Dict containing ent1 objects as keys and feature vectors (side information) as values.

			Returns
			-------
		"""
		self.name = name
		self.ent1 = ent1
		self.dtype = type
		self.data = {} # data = {idx:[features]}
		missing = 0
		l = len(data.values()[0])
		for i in ent1:
			try:
				self.data[ent1[i]] = data[i]
			except:
				self.data[ent1[i]] = [0]*l
				missing += 1
		print ("Missing: ", missing)
		for i in self.data.items():
			assert len(i[1]) == len(self.data.values()[0])

	@dispatch(str,  NX.Entity, np.ndarray)
	def __init__(self, name, ent1, data):
		"""
			One of the alternative constructors for the SideInfo class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			data: numpy.ndarray
				Numpy array that contains the side information. It has shape (ent1 obj, feature length), similarly to a scikit-learn feature vector.

			Returns
			-------
		"""
		self.name = name
		self.ent1 = ent1
		self.dtype = type
		self.data = {} # data = {idx:[features]}
		missing = 0
		assert len(ent1) == data.shape[0], "ERROR: data.shape[0] and len(ent1) do not match. You should provide one side info vector for each object in the entity."
		l = data.shape[1]
		for i in ent1:
			try:
				self.data[ent1[i]] = data[i]
			except:
				self.data[ent1[i]] = [0]*l
				missing += 1
		print ("Missing: ", missing)

	@dispatch(str,  NX.Entity, coo_matrix)
	def __init__(self, name, ent1, data):
		"""
			One of the alternative constructors for the SideInfo class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			data: scipy.sparse.coo_matrix
				Scipy coo_matrix that contains the side information. It has shape (ent1 obj, feature length), similarly to a scikit-learn feature vector. It can be sparse, but currently the sparsity during mini batching is NOT supported.

			Returns
			-------
		"""
		self.name = name
		self.ent1 = ent1
		self.dtype = type
		d = {}
		l = data.shape[1]
		data = data.toarray()
		print(data.shape)
		i = 0
		while i < len(data.shape[1]):
			d[i] = data[i,:]
		data = d
		self.data = {} # data = {idx:[features]}
		missing = 0
		for i in ent1:
			print (i, ent1[i], data[i])
			try:
				self.data[ent1[i]] = data[i][:100]
			except:
				self.data[ent1[i]] = [0]*100
				missing += 1
		print ("Missing: ", missing)
		#for i in self.data.items():
		#	assert len(i[1]) == len(self.data.values()[0])

	@dispatch( str)
	def __init__(self, path):
		"""
		This constructor reads a serialized (SideInfo.dump()) SideInfo object.			
			Parameters
			----------
			name: str
				Path to the serialized SideInfo object.

			Returns
			-------
		"""

		print( "Loading %s..." % path)
		start = time.time()
		store = cPickle.load(open(path))
		stop = time.time()

		try:
			store["name"]
			store["ent1"]
			store["data"]
		except:
			print( "ERROR: wrong format, check file content")
			exit(1)
		self.name= store["name"]
		self.ent1 = store["ent1"]
		self.data = store["data"]
		print ("Done in %.2fs." % (stop-start))

	def normalize(self):
		"""
		Method that standardizes the matrix with the formula x' = (x - mu)/s, where mu is the mean and s is the standard deviation.

		Returns
		-------
		None
		"""

		l = []
		for i in self.data.values():
			l += list(i[1])
		#print l
		print( len(l))
		mu = np.mean(l)
		s = np.std(l)
		print ("mu = %f, s= %f" % (mu, s))
		for i in self.data.items():
			self.data[i[0]] = [i[1][0], (i[1][1]-mu)/s]

	def __len__(self):
		return len(self.data)

	def __getitem__(self, x):
		return self.data[x]

	def dump(self, path=None):
		"""
		Method that serializes the SideInfo storing it at the selected path.

		path: str
			Destination path for the serialized file

		Returns
		-------
		None
		"""
		if path == None:
			print ("Storing...")
			start = time.time()
			store = {"name":self.name, "ent1":self.ent1, "data":self.data}
			cPickle.dump(store, open("marshalled/"+self.name+".side.nx", "w"))
			stop = time.time()
			print ("Done in %.2fs." % (stop-start))
		else:
			print ("Storing...")
			start = time.time()
			store = {"name":self.name, "ent1":self.ent1, "data":self.data}
			cPickle.dump(store, open(path, "w"))
			stop = time.time()
			print ("Done in %.2fs." % (stop-start))


class DataMatrix(object):

	"""
	The input "data" format should be: {(ent1, ent2): value} for all the observed elements in the matrix.

	The format in which the data is stored in the DataMatrix object is the following: 
	featsHT = {domain1Name_numeric : [ numpy16_domain2Names_numeric, numpyX_labels ]}

	"""

	@dispatch(str, NX.Entity,  NX.Entity, dict, type)
	def __init__(self, name:str, ent1:NX.Entity, ent2:NX.Entity, data:dict, dtype:type) :
		"""One of the alternative constructors for the DataMatrix class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			ent2: Entity
				Entity object representing the object on the dimension 1
			data: dict  {(ent1, ent2): value}
				Hash table containing the (sparse) elements and in the matrix describing the relation. The input "data" format should be: {(ent1, ent2): value} for all the observed elements in the matrix.
			dtype: numpy.dtype
				The smallest possible type that could be used to store the elements of the matrix (e.g. np.int16)

			Returns
			-------
		"""
		self.name = name
		self.ent1 = ent1
		self.ent2 = ent2
		self.dtype = type
		print( "Building features for matrix %s..." % name)
		self.data, self.size = buildPytorchFeatsHT(data, ent1, ent2, dtype)
		#print (list(self.data.items())[:3])
		print ("Size: ", self.size)

	@dispatch(str, NX.Entity,  NX.Entity, coo_matrix)
	def __init__(self, name:str, ent1:NX.Entity, ent2:NX.Entity, data:coo_matrix) :
		"""One of the alternative constructors for the DataMatrix class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			ent2: Entity
				Entity object representing the object on the dimension 1
			data: coo_matrix
				scipy.sparse.coo_matrix containing the sparse elements and in the matrix describing the relation.

			Returns
			-------
		"""
		self.name = name
		self.ent1 = ent1
		self.ent2 = ent2
		self.dtype = type
		print( "Building features for matrix %s..." % name)
		d = {}
		data = data.todok()
		for i in data.items():
			d[i[0]] = i[1]
		self.data, self.size = buildPytorchFeatsHT(d, ent1, ent2, data.dtype)
		#print (list(self.data.items())[:3])
		print ("Size: ", self.size)


	@dispatch(str, NX.Entity, NX.Entity, np.ndarray)
	def __init__(self, name:str, ent1:NX.Entity, ent2:NX.Entity, data:np.ndarray) :
		"""
			One of the alternative constructors for the DataMatrix class.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			ent1 : Entity 
				Entity object representing the object on the dimension 0
			ent2: Entity
				Entity object representing the object on the dimension 1
			data: numpy.ndarray (matrix)
				Numpy matrix containing the (dense) describing the relation between ent1 and en2. 

			Returns
			-------
				DataMatrix object
		"""
		self.name = name
		self.ent1 = ent1
		self.ent2 = ent2
		assert len(ent1) == data.shape[0], "ERROR: ent1 len is "+str(len(ent1))+" but data.shape[0] is"+str(data.shape[0])
		assert len(ent2) == data.shape[1], "ERROR: ent2 len is "+str(len(ent2))+" but data.shape[0] is"+str(data.shape[0])

		print( "Building features for matrix %s..." % name)
		self.data, self.size = buildPytorchFeatsHTfromNumpy(data, self.ent1, self.ent2)
		#print (list(self.data.items())[:3])
		print ("Size: ", self.size)

	@dispatch(str, np.ndarray)
	def __init__(self, name:str, data:np.ndarray, dtype:type) :
		"""Simplest possible constructor for the DataMatrix class. Entities are inferred.
			
			Parameters
			----------
			name: str
				Name of the data matrix
			data: numpy.ndarray (matrix)
				Numpy matrix containing the (dense) describing the relation between ent1 and en2. 

			Returns
			-------
		"""

		self.name = name
		self.ent1 = NX.Entity(self.name+"_0", list(range(data.shape[0])), dtype = dtype)	
		self.ent2 = NX.Entity(self.name+"_1", list(range(data.shape[1])), dtype = dtype)
		print( "Building features for matrix %s..." % name)
		self.data, self.size = buildPytorchFeatsHTfromNumpy(data, self.ent1, self.ent2)
		#print (list(self.data.items())[:3])
		print ("Size: ", self.size)

	def size(self):
		"""
		Function that return the size of the relation (number of elements in the matrix).

		Returns
		-------
			Size of the relation in the DataMatrix object
		"""
		size = 0
		for i in data.items():
			size += len(i[1][0])
		print ("Size: ", size)
		return size

	@dispatch( str)
	def __init__(self, path):
		"""Constructor that reads the DataMatrix from a previously serialized DataMatrix object.
			
			Parameters
			----------
			path: str
				Path of the serialized DataMatrix
			
			Returns
			-------
					"""

		print( "Loading %s..." % path)
		start = time.time()
		store = cPickle.load(open(path))
		stop = time.time()

		try:
			store["name"]
			store["ent1"]
			store["ent2"]
			store["data"]
		except:
			print ("ERROR: wrong format, check file content")
			exit(1)
		self.name= store["name"]
		self.ent1 = store["ent1"]
		self.ent2 = store["ent2"]
		self.data = store["data"]
		print ("Done in %.2fs." % (stop-start))

	def standardize(self):
		"""
		Method that standardizes the matrix with the formula x' = (x - mu)/s, where mu is the mean and s is the standard deviation.

		Returns
		-------
		None
		"""
		l = []
		for i in self.data.values():
			l += list(i[1])
		#print l[:100]
		#print len(l)
		mu = np.mean(l)
		s = np.std(l)
		print ("mu = %f, s= %f" % (mu, s))
		tmp = []
		for i in self.data.items():
			res = (i[1][1]-mu)/s
			#print res
			#raw_input()
			self.data[i[0]] = [i[1][0], res]
			tmp += res.tolist()
		#print len(tmp)
		#print tmp[:100]
		tmu = np.mean(tmp)
		ts = np.std(tmp)
		#print tmu, ts
		assert abs(tmu) < 0.0001 
		assert abs(ts-1) < 0.0001
		
	def toHashTable(self)-> dict:
		"""
		Method that returns an hash table (dict) containing the DataMatrix data.
		
		Returns
		-------
			dict
		"""

		db = {}
		for p1 in self.data.items():
			#print p1
			i = 0
			assert len(p1[1][0]) == len(p1[1][1])
			while i < len(p1[1][0]):
				db[tuple(sorted([self.ent1[p1[0]], self.ent2[int(p1[1][0][i])]]))] = p1[1][1][i]
				i+=1
		print( "Found %d entries" % len(db))
		#print db.items()[:10]
		return db

	def dump(self, path= None):
		if path == None:
			print ("Storing...")
			start = time.time()
			store = {"name":self.name, "ent1":self.ent1, "ent2":self.ent2, "data":self.data}
			cPickle.dump(store, open("marshalled/"+self.name+".nx", "w"))
			stop = time.time()
			print ("Done in %.2fs." % (stop-start))
		else:
			print ("Storing...")
			start = time.time()
			store = {"name":self.name, "ent1":self.ent1, "data":self.data}
			cPickle.dump(store, open(path, "w"))
			stop = time.time()
			print ("Done in %.2fs." % (stop-start))

@dispatch(np.ndarray, NX.Entity, NX.Entity)	
def buildPytorchFeatsHTfromNumpy(data:np.ndarray, domain1:NX.Entity, domain2:NX.Entity) -> (dict, int):
	""" This function produces the data structure that is internally used to pass training data to the wrapper.fit function. This is now transparent to the user.
		
		Parameters
		----------
		data: numpy.ndarray
			Numpy matrix containing the matrix that represents the relation between ent1 and ent2
		domain1 : Entity
			Entity1
		domain2 : Entity
			Entity2

		Returns
		-------
		Dict internally used to feed the data to the NX.Wrapper object.
		The format is the following
		featsHT = {domain1Name_numeric : [ numpy16_domain2Names_numeric, numpyX_labels ]}
	"""

	size = 0
	x = {}
	i = 0
	while i < len(domain1):
		x[i] = [[],[]]
		i+=1
	if domain1 == domain2:
		print( " *** identified as self relation.")
		selfRelation = True
	else:
		print (" *** identified as asymmetric relation.")
		selfRelation = False
	for i in range(0, len(domain1)):
		for j in range(0, len(domain2)):
			if selfRelation:
				try:
					tmp = tuple(sorted([i,j])) #probably dangerous here
				except:
					continue
			else:
				tmp = tuple([i,j])
			x[tmp[0]][0].append(tmp[1])
			x[tmp[0]][1].append(data[i,j])
	xf = {}

	for i in x.items():
		size += len(i[1][0])
		xf[i[0]] = [np.array(i[1][0], dtype=domain1.dtype), np.array(i[1][1], dtype=data.dtype)]
	return xf, size

@dispatch(dict, NX.Entity, NX.Entity, np.dtype)
def buildPytorchFeatsHT(data:dict, domain1:NX.Entity, domain2:NX.Entity, relDtype:type)-> (dict, int):
	""" This function produces the data structure that is internally used to pass training data to the wrapper.fit function. This is now transparent to the user.
		
		Parameters
		----------
		data: dict {(ent1[i], ent2[j]): value[i,j]}

			Dict in the following format: {(ent1[i], ent2[j]): value[i,j]} containing the matrix that represents the relation between ent1 and ent2. This format can be used to input sparse matrices
		domain1 : Entity
			Entity1
		domain2 : Entity
			Entity2
		relDtype : numpy.dtype
			The smallest np.dtype sufficient to represent the values in the matrix. 

		Returns
		-------
		Dict internally used to feed the data to the NX.Wrapper object.
		The format is the following
		featsHT = {domain1Name_numeric : [ numpy16_domain2Names_numeric, numpyX_labels ]}
	
	"""

	size = 0
	x = {}
	i = 0
	while i < len(domain1):
		x[i] = [[],[]]
		i+=1
	if domain1 == domain2:
		print( " *** identified as self relation.")
		selfRelation = True
	else:
		print (" *** identified as asymmetric relation.")
		selfRelation = False
	for i in data.items():
		if selfRelation:
			try:
				tmp = tuple(sorted([domain1[i[0][0]], domain2[i[0][1]]])) #probably dangerous here
			except:
				continue
		else:
			tmp = tuple([domain1[i[0][0]], domain2[i[0][1]]])
		x[tmp[0]][0].append(tmp[1])
		x[tmp[0]][1].append(i[1])
	xf = {}

	for i in x.items():
		size += len(i[1][0])
		xf[i[0]] = [np.array(i[1][0], dtype=domain1.dtype), np.array(i[1][1], dtype=relDtype)]
	return xf, size





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pytorchDatasetUtils.py
#  
#  Copyright 2018 Daniele Raimondi <daniele.raimondi@vub.be>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import pickle as cPickle
import time
import torch as t
import numpy as np
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader

class PredictionDatasetSide(Dataset):
	"""
	:meta private:
	"""
	def __init__(self, x, sx1, sx2):		
		#if sx1 != None and len(sx1) > 0:
		#	assert len(x) == len(sx1) 
		#if sx2 != None and len(sx2) > 0:
		#	assert len(sx2) == len(x)
		self.x = x # [(i,j), (i,k), ...]
		self.sx1 = sx1
		self.sx2 = sx2
				
	def __getitem__(self, idx): 
		"""
		:meta private:
		"""

		tmp = self.x[idx]
		sx1 = []
		sx2 = []
		if self.sx1 != None:
			sx1 = self.sx1[tmp[0]]
		if self.sx2 != None:
			sx2 = self.sx2[tmp[1]]
		return tmp[0], tmp[1], sx1, sx2		
	
	def __len__(self):
		return len(self.x)


class PredictionDataset(Dataset):
    
	def __init__(self, x, label = True):		
		self.x = x # [(i,j), (i,k), ...]
				
	def __getitem__(self, idx): 
		tmp = self.x[idx]
		return tmp[0], tmp[1]
		
	def __len__(self):
		return len(self.x)

class SideDataset(Dataset):
	def __init__(self, side):
		assert type(side) == {}
		self.side = side
		self.estSize = len(self) * len(self.values()[0])
	
	def __getitem__(self, idx):
		return self.side[idx]

	def __len__(self):
		return len(self.side)

class SubDataset(Dataset):
	"""
	Within the NNwrapper, during training, batches need to be rapidly provided for all the MetaRelations in the ERgraph and for each Relation in every MetaRelation. To do so, the NNwrapper.processDatasets function builds an internal Dataset structure that mimicks the structure of the input ERgraph. In this case, MetaDataset correspond to MetaRelation, and each Relation in a MetaRelation is represendet by a SubDataset in the corresponding MetaDataset.

	Nevertheless, this is internal and it is transparent to the user.
	:meta private:
	"""
	def __init__(self, xht, typep="binary"):		
		"""
		Constructor method for the SubDataset class. It puts in a pytorch-friendly structure the matrix corresponding to a target Relation, by transforming its DataMatrix into a pytorch Dataset.

		Parameters
		----------
		xht : dict
			Dict used to represent the matrix/relation data within a DataMatrix object
		type : str 
			String specifying the type of the prediction. It must be "regression" or "binary".

		Returns
		-------
		
		"""
		#print type(xht)
		assert type(xht) == dict
		self.xht = xht #xht = {p1:[(positions),(values)]}
		empty = 0
		for i in self.xht.items():
			if len(i[1][0]) == 0:
				#print i[0]
				empty += 1
				#print i[1]
		
		print ("Empty rows: ",empty) 
		#raw_input()
		self.estSize = self.countInstances()
		self.type = typep	
		self.balance = self.countBalance()
		
	def countBalance(self):
		if self.type != "binary":
			return "regression"
			#raise Exception("CAlling count balance on regression problem!!!")
		r = [0,0]
		for i in self.xht.values():
			tmp = sum(i[1])
			r[0] += tmp
			r[1] += len(i[1])-tmp
		return r
		
	def countInstances(self):
		r = 0
		for i in self.xht.values():
			r += len(i[0])
		return r
		
	def __getitem__(self, idx):
		tmp = self.xht[idx]
		return tmp
		#return (np.array(tmp[0], dtype=np.int16), np.array(tmp[1], dtype=np.int8))
	
	@staticmethod
	def load(name):
		tmp = cPickle.load(open(name))
		vt = None
		if "binary" in name:
			vt = "binary"
		elif "regression" in name:
			vt = "regression"
		return SubDataset(tmp, typep=vt)
	
	def dump(self, name):
		print( "Dumping...")
		t1 = time.time()
		cPickle.dump(self.xht, open(name, "w"))
		t2 = time.time()
		print ("Stored in: %s (%.2fs)" % ( name, t2-t1))
		
	def __len__(self):
		return len(self.xht)		

class MetaDataset(Dataset): ###########cercato di ottimizzre lo spazio con numpy ma al collate arriva roba strana, da fixare !!!!!!!!!!#######################
	"""
	Class that represents the MetaRelations in the NNwrapper internal Dataset-based version of the ERgraph used for allowing a fast and consistent multi-task mini batching. Each MetaDataset can contain many SubDatasets, and when asked it provides a minibatch sampling from all of them in parallel.
	"""
	def __init__(self, datasetList, domain1, domain2, name, ignore_index, side1 = None, side2 = None):
		"""
		Constructor method for the MetaDataset class. It puts in a pytorch-friendly structure the data corresponding to a target MetaRelation, by storing several SubDataset (each corresponding to a Relation/DataMatrix/matrix).

		Parameters
		----------
		datasetList : list of SubDatasets
			List of Subdatasets. Each SubDataset corresponds to a Relation. The MetaDataset thus corresponds to a MetaRelation.
		domain1 : NX.Entity
			First entity involved in this MetaRelation (all the Relations in it are between the same entities)
		domain2 : NX.Entity
			Second entity involved in this list of relations (MetaRelation).
		name: str
			Name of the corresponding MetaRelation
		ignore_index: int
			Value corresponding to missing values. Used to allow fast runs on GPUs and minibatching even with different percentages of missing values among the Relations/SubDatasets in the same MetaRelation/MetaDataset.

		Returns
		-------
		
		"""
		self.name = name
		self.ignore_index = ignore_index
		self.side1 = side1
		self.side2 = side2
		self.datasetList = datasetList
		self.domain1 = domain1
		self.domain2 = domain2

	def getTypes(self):
		r = []
		for i in self.datasetList:
			r.append(i.type)
		return r

	def countBalance(self):
		r = []
		for i in self.datasetList:
			r.append(i.balance)
		return r

	def countInstances(self):
		r = []
		for i in self.datasetList:
			r.append(i.estSize)
		return r

	def getEstSize(self):
		return sum(self.countInstances())

	def getEstBatchSizeForXsamples(self, targetDomain1, samplesPerBatch):
		assert samplesPerBatch > 0
		perc = samplesPerBatch / float(len(targetDomain1))
		res = int(max(1, perc * float(len(self.domain1))))
		print (" Foreseen batch size: ", res)
		return res

	def getEstBatchSizeForXsamples2(self, numSamples): #future daniele, pay attention to this, may be working bady
		assert numSamples > 10
		tot = self.getEstSize()/len(self.domain2)
		res = max(1,tot / numSamples)
		print( " Foreseen batch size: ", res)
		return res

	def __getitem__(self, idx):
		tmp = []
		for d in self.datasetList:
			tmp.append(d[idx])
		tmp = self.mergeDataSimple(tmp, idx)
		#print tmp
		#raw_input()
		return tmp
	
	def mergeDataSimple(self, v, idx):
		x1 = []
		x2 = []
		y = []
		xside1 = []
		xside2 = []
		#da sistemare qui sotto
		for count, ds in enumerate(v):
			#print count
			assert len(ds[0]) == len(ds[1]) and len(ds) == 2
			tmp1 = np.array([idx]*len(ds[0]), dtype=np.int32) #BEWARE! you might have more instances TODO
			#print "x:", tmp1.shape, len(ds[0])
			x1.append(tmp1)
			x2.append(ds[0])
			tmp = np.ones((len(ds[0]), len(v)))* self.ignore_index
			tmp[:,count] = ds[1]

			if self.side1 != None and len(ds[0]) > 0:
				#print self.side1[idx]
				#print len(ds[0])
				for s in xrange(0, len(ds[0])):
					xside1.append( self.side1[idx])
				
			if self.side2 != None and len(ds[0]) > 0:
				tmp1 = []
				for s in ds[0]:
					xside2.append(self.side2[s])
			y.append(tmp)
			#print tmp.shape
		if self.side1 != None and len(xside1) > 0:
			xside1 = np.vstack(xside1)
		else:
			xside1 = []
		if self.side2 != None and len(xside2) > 0:
			xside2 = np.vstack(xside2)
		else:
			xside2 = []
		#print xside2
		if self.side1 != None or self.side2 != None:
			#print len(xside1), len(xside2)
			#print xside1.shape, xside2.shape
			return np.hstack(x1), np.hstack(x2), np.vstack(y), xside1, xside2
		else:		
			return np.hstack(x1), np.hstack(x2), np.vstack(y)

	def __len__(self):
		return len(self.domain1)	





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pytorchLossUtils.py
#  
#  Copyright 2018 Daniele Raimondi <daniele.raimondi@vub.be>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import time
import torch as t
from torch.autograd import Variable

class FocalLoss(t.nn.Module):
	"""
	Implementation of the FocalLoss, which is used as loss for heavily unbalanced binary predictions. It is bult as a convetional pytorch module.
	"""
	def __init__(self, alpha=1, gamma=2, logits=True, reduction="sum"):
		"""
			Constructor for the FocalLoss.

			Parameters
			----------
			alpha : int
				Parameter of the loss
			gamma : int
				Parameter of the loss
			logits : bool
				Uses logits if True
		"""
		super(FocalLoss, self).__init__()
		self.alpha = alpha
		self.gamma = gamma
		self.logits = logits
		self.reduction = reduction

	def forward(self, inputs, targets):
		"""
		:meta private:
		"""

		if self.logits:
			BCE_loss = t.nn.functional.binary_cross_entropy_with_logits(inputs, targets, reduce=False)
		else:
			BCE_loss = t.nn.functional.binary_cross_entropy(inputs, targets, reduce=False)
		pt = t.exp(-BCE_loss)
		F_loss = self.alpha * (1-pt)**self.gamma * BCE_loss

		if self.reduction == "mean":
			return t.mean(F_loss)
		elif self.reduction == "sum":
			return t.sum(F_loss)
		else:
			return F_loss



class LossWrapper(t.nn.Module):

	"""
	Class that wraps any pytorch loss allowing for ignore index. In the Matrix Factorization context it may be useful to define a value indicating missing values even when performing a regression, for example if the goal is to predict a sparsely observed real-valued matrix.
	"""

	def __init__(self, loss:t.nn.Module, type:str, ignore_index:int):
		"""
		Constructor for the wrapper.

		Parameters
		----------
		loss : t.nn.Module
			The argument can be any pytorch compatible loss functioni
		type:
			Specifies wheter is a regression or a binay prediction (deprecate?)
		ignore_index : int
			Specifies which value should be ignored while computing the loss, to allow for the presence of missing values in the matrix/relation.

		Returns
		-------
		"""
		super(LossWrapper, self).__init__()
		self.loss = loss
		self.ignore_index = ignore_index
		self.type = type
	
	def __call__(self, input, target):
		"""
		Function defining the forward pass for this wrapper. It implements the ignore_index filtering and then it calls the actual self.loss on the remaining values.

		Parameters
		----------
		input : t.nn.Tensor
			Pytorch tensor containing the predicted values
		target : t.nn.Tensor
			Pytorch tensor containing the target values

		Returns
		-------
		Loss score computed only for the target values that are not equal to self.ignore_index.
		"""
		input = input.view(-1)
		target = target.view(-1)
		if self.ignore_index != None:
			mask = target.ne(self.ignore_index)
			input = t.masked_select(input, mask)
			target = t.masked_select(target, mask)
		#t2 = time.time()
		
		r = self.loss(input, target)
		#t3 = time.time()
		#print t2-t1, t3-t2
		#raw_input()
		return r




import pickle as cPickle
import time
import torch as t
import numpy as np
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader

class MetaLoader:

	def __init__(self, loaders):
		self.loaders = loaders

	def __iter__(self):
		return MetaLoaderIter(self)

class MetaLoaderIter:

	def __init__(self, metaLoader):
		self.loadersIter = self.createIterators(metaLoader.loaders)
		self.counter = []
	
	def createIterators(self, loaders):
		loadersIterators = []
		i = 0
		while i < len(loaders):
			loadersIterators.append(iter(loaders[i]))
			i+=1
		return loadersIterators

	def __iter__(self):
		return self

	def __next__(self):
		r = []
		for i in self.loadersIter:
			r.append(next(i))
		return r

	next = __next__





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  ppiPred.py
#  
#  Copyright 2017 Daniele Raimondi <eddiewrc@alice.it>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import pickle as cPickle
import os, sys, gc, marshal, copy, random, time, socket, math
import scipy.stats as st
import NXTfusion.NXFeaturesConstruction as FC
import NXTfusion.NXMetaLoaderUtils as MLU
import NXTfusion.NXCollateUtils as Collate
from sys import stdout
import numpy as np
import torch as t
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader
import NXTfusion.NXLosses as L
import NXTfusion.NXDatasetUtils as D
import NXTfusion.NXTfusion as NX
from NXTfusion.Logger import MetaLogger

#np.seterr(all="raise")

class NNwrapper():
	"""
	Class that wraps a t.nn.Module (pytorch module) and uses scikit-learn-like methods such as .fit() and .predict() to train and test it.
	"""

	def __init__(self, model, dev, ignore_index, initialEpoch=0, nworkers = 0):
		"""
		Constructor for the NNWrapper class, which facilitates and standardizes the training of pytorch neural networks.

		Parameters
		----------
		model : t.nn.Module 
			The pytorch Neural Network that should be trained or tested.
		def : t.device
			The device on which the model should run. E.g. t.device("cuda") or t.device("cpu:0")
		ignore_index : int
			The ignore index value that will be used to mark "missing values" and "N/A" on partially observed matrices, in order to let the corresponding loss ignore those instances.
		

		Returns
		-------
		"""
		self.model = model
		self.ignore_index = ignore_index
		self.initialEpoch = initialEpoch +1
		self.logger = None
		self.save_model_every = 5
		self.nworkers = nworkers
		self.model.to(dev)
		self.dev = dev
		print( "***** Model device: ", self.dev)

	def processDatasets(self, DS:list):

		""" 
		This method takes the external Entity Relation graph representation, in the form of one MetaRelation at a time and converts it
		into lower level data structures to be used within the wrapper, 
		creating a MetaDataset structure from the ER representation passed as input. This structure mimics the ERgraph, but it's suitable for efficient multi-task mini batching during training.
		This function is used internally by the NNwrapper and does not need to be called by the user.
		

		Parameters
		----------
		DS : MetaRelation 
			
		Returns
		-------
		DS : MetaRelation
			The original MetaRelation without the data matrices, in an attempt to save space. (still have to run benchmarks on it)
		datasets : list of SubDataset
		losses : list of losses
		refSize : size of the target matrix (to be removed)
		:meta private:
		"""

		datasets = []
		losses = []
		refSize = None
		print ("Loading datasets...")
		for i, ds in enumerate(DS):
			print ("Working on ds %s..." % ds["name"])
			#if ds["type"] == "DS": #realDataset
			datasets.append(D.SubDataset(ds["data"].data))
			DS[i].__delitem__("data")		
			if ds["name"] == "target":
				refSize = datasets[-1].estSize
			losses.append(ds["loss"])
		print ("Done.")
		gc.collect()
		#assert refSize != None
		return DS, datasets, losses, refSize

	def saveModel(self, e:int):
		"""
		Method that stores the trained model at a certain iteration. Used internally.

		Parameters
		----------
		e : int
			Epoch number. The model is automatically saved using the model name and the epoch number using t.save function.

		"""
		if e == True or e % self.save_model_every == 0:
			t.save(self.model, self.model.name+".iter_"+str(e)+".t")	
			print ("Store model ", e)			
			stdout.flush()

	def countParams(self, parameters:list) -> int:
		"""
		Method that counts the number of trainable parameters in the model.

		Parameters
		----------
		parameters : iterable
			The iterable containtaining the pytorch model parameters.

		Returns
		-------
		Number of parameters (int)
		
		"""

		p = []
		for i in parameters:
			p+= list(i.data.cpu().numpy().flat)
		print ('Number of parameters=',len(p))
		return len(p)

	def getRelationWeights(self, relData, datasets, refSize):
		"""

		:meta private:
		"""
		relationWeight = []
		for i, ds in enumerate(relData):
			if type(ds["relationWeight"]) != str:
				relationWeight.append(ds["relationWeight"])
			elif ds["relationWeight"] == "relativeToTarget":
				w = refSize/float(datasets[i].estSize)
				relationWeight.append(w)
			print ("Rel %d, name: %s, type: %s, loss: %f" % (i, ds["name"], str(ds["relationWeight"]), relationWeight[-1]))
		tmp = t.tensor(relationWeight, dtype=t.float, device=self.dev)
		return tmp

	def computeLosses(self, y, yp, losses, relationData, weightRelations):
		"""
		This function computes the losses for the entire ER graph, by iterating through them.  Used internally.

		Parameters
		----------
		y : t.tensor
			Pytorch tensor containing the labels
		yp : t.tensor
			Pytorch tensor containing the predictions
		losses : list 
			list of losses (LossWrapper or t.nn.Module)
		relationData : list
			list of MetaRelations
		weightRelations : list
			list of weights associated to each loss

		Returns
		-------
		loss : real
			total loss
		tmpLoss : list
			list containing the losses associated to each Relation
		:meta private:
		"""
		tmpLoss = []
		assert len(losses) == len(relationData) == len(weightRelations)
		for ci, l in enumerate(losses):
			if l.type == "binary":
				assert l.type == relationData[ci]["task"]
				tmpLabel = y[:,ci:ci+1,].squeeze()
			elif l.type == "regression":
				assert l.type == relationData[ci]["task"]
				tmpLabel = y[:,ci:ci+1,].squeeze()
			else:
				raise Exception("unrecognized")
			tmpLoss.append(weightRelations[ci] * losses[ci](yp[:,ci:ci+1], tmpLabel))
		#print tmpLoss
		return sum(tmpLoss), tmpLoss 	

	def printBatchesLog(self, rel, e, bi, errTotOld, errTot, totLen, epochTime, loadTime, forwTime, LossTime, start, batch_size, mute = True):

		""" This class simplifies the live logging of the batches. If muted, it will only signal excessively long loading times.

		Parameters
		----------
		TODO

		Returns
		-------
		:meta private:
		"""
		if mute:
			if loadTime > 0.5:
				print( "WARNING: loadTime %3.2fs !" % loadTime)
		else:
			sys.stdout.write("rel: %d, e: %d, b: %d/%d, %3.2f%% (T:%.3fs, e:%.3fs, l:%.3fs, f:%.3fs, b:%.3fs)\n" % (rel, e, bi, totLen, 100*(bi/float(totLen)), time.time() - start, epochTime, loadTime, forwTime, LossTime, )	)
			sys.stdout.flush()	# T: total, e: epochtime, l: load, f: forward, b: back

	def printLossLog(self, e, datasetsList, errors, start, end, errTotOld, errTot):
		"""
		This function prints the logs at each epoch. Used to make the code less cumbersome to read.

		Parameters
		----------
		TODO

		Returns
		-------
		
		:meta private:
		"""
		sys.stdout.write("\n epoch %d,"% e)
		percRed = ((errTotOld-errTot)/float(errTotOld))*100.0	
		#print errors, len(errors), len(tmpLoss)
		sys.stdout.write(" ERRORTOT: %f (%fs) %2.3f%%\n" % (errTot, end-start, percRed))
		for i, rel in enumerate(errors):	
			sys.stdout.write("Relation: %d\n" % i)
			for li, l in enumerate(rel):
				sys.stdout.write("\t Mat: %s: error: %.2f\n" % (datasetsList[i][li]["name"], float(rel[li])))
		sys.stdout.write("\n")
		sys.stdout.flush()
	
	def buildTensorboardLog(self, errTot, lossScores, relationList, te):
		"""
		:meta private:
		"""
		if self.logger == None:
			print (" WARNING: Logger not initializated, skipping")
			return
		info = {"errTot":errTot}
		for r in relationList:
			for i, l in enumerate(lossScores):
				info[r[i]["name"]] = l
		info["timePerEpoch"] = te

		for tag, value in info.items():
			logger.scalar_summary(tag, value, step+1)

		# (2) Log values and gradients of the parameters (histogram)
		for tag, value in self.model.named_parameters():
			#print
			#print tag, value, value.grad					
			tag = tag.replace('.', '/')
			logger.histo_summary(tag, to_np(value), step+1)
			logger.histo_summary(tag+'/grad', to_np(value.grad), step+1)

		return info
		
	def fit(self, relationList, epochs = 100, batch_size=500, save_model_every=10, LOG=False, MUTE = True):
		"""
		Function that performs the training of the wrapped pytorch model. 
		It is analogous to scikit-learn .fit() method.


		Parameters
		----------
		relationList : ERgraph
		epochs : int
			Number of epochs
		batch_size : int
			batch size during training
		save_model_every : int
			Stores the model every int epochs

		Returns
		-------
		

		"""
		print ("self.dev: ", self.dev)
		self.model.train()
		print ("Training mode: ", self.model.training)
		FAST = False 
		self.save_model_every = save_model_every
		if LOG:
			self.logger = MetaLogger(self.model, port = 6001)
		########DATASET###########
		print ("Processing %d relations " % len(relationList))
		datasetsList = []
		lossesList = []
		weightRelations = []
		refSize = None
		for i, relData in enumerate(relationList):
			r, datasets, losses, tmpRefSize = self.processDatasets(relData) #relations, pytorchDatasets, losses
			if tmpRefSize != None:
				assert refSize == None
				refSize = tmpRefSize
			datasetsList.append(D.MetaDataset(datasets, r[0]["domain1"], r[0]["domain2"], relData.name, self.ignore_index, relData.side1, relData.side2))
			weightRelations.append(self.getRelationWeights(relData, datasets, refSize))
			relationList[i] = r
			lossesList.append(losses)
			gc.collect()
		#######MODEL##############		
		parameters = self.model.getParameters()
		self.countParams(parameters)	
		self.model.train()
		print( "Training mode: ", self.model.training)

		print ("Start training")
		########OPTIMIZER##########	
		self.learning_rate = 1e-3 
		if  next(self.model.parameters()).is_cuda: 
			assert next(self.model.parameters()).is_cuda and "cuda" in self.dev
		optimizer = t.optim.Adam(parameters, lr=self.learning_rate, weight_decay=0.001)
		scheduler = t.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.5, patience=1, verbose=True, threshold=0.00001, threshold_mode='rel', cooldown=0, min_lr=0, eps=1e-08)
		
		########DATALOADER#########
		print( "Creating dataloaders for %s..." % datasetsList)
		loaders = []
		for ds in datasetsList:
			loaders.append(DataLoader(ds, batch_size=ds.getEstBatchSizeForXsamples(relationList[0][0]["domain1"], 100), shuffle=True, sampler=None, num_workers=self.nworkers, collate_fn=Collate.metaCollate, pin_memory=False))
		metaLoader = MLU.MetaLoader(loaders)	
		e = self.initialEpoch
		assert len(datasetsList) == len(loaders) == len(metaLoader.loaders)
		errTot = 0
		###############TRAINING ITERATIONS#############
		while e < epochs + self.initialEpoch:		
			errTotOld = errTot	
			errTot = 0
			errors = []
			for r in relationList:
				errors.append([0]*len(r))
			c = 0	
			start = time.time()		
			bi = []
			for b in datasetsList:
				bi.append(0)
			tLoad1 = time.time()	# start epoch time
			for sample in metaLoader:
				tLoad2 = time.time() # end load time, start forward time
				lossTot = 0
				optimizer.zero_grad()	
				
 				#D.printSampleSizes(sample, relationList)
				for r, rel in enumerate(sample):
					#if self.CUDA:
					#	t.cuda.empty_cache()
					bi[r] += loaders[r].batch_size
					if len(rel) > 3: #if side information is present
						x1, x2, y, side1, side2 = rel
						#	print side1, side2
						x1 = t.tensor(x1, dtype=t.long, device=self.dev)
						x2 = t.tensor(x2, dtype=t.long, device=self.dev)
						y = t.tensor(y, dtype=t.float, device=self.dev)
						if type(side1) != type(None):
							#print "side1 not none"
							#print side1
							side1 = t.tensor(side1, dtype=t.float, device=self.dev)
						if type(side2) != type(None):
							#print side2.shape
							side2 = t.tensor(side2, dtype=t.float, device=self.dev).squeeze()
							#print "side2 not none"
					else:
						x1, x2, y = rel
						x1 = t.tensor(x1, dtype=t.long, device=self.dev)
						x2 = t.tensor(x2, dtype=t.long, device=self.dev)
						y = t.tensor(y, dtype=t.float, device=self.dev)
						#print y.tolist(), r
						#raw_input()
					#print "Rel: %d, size: %s" % (r, str(x1.size()))
					#print "Rel %d, batch size %s" % (r, x1.size())	
					tForw1 = time.time()
					assert datasetsList[r].name == relationList[r].name
					if len(rel) > 3:
						yp = self.model.forward(relationList[r].name, x1, x2, NX.SafeVariable(side1), NX.SafeVariable(side2))
					else:
						yp = self.model.forward(relationList[r].name, x1, x2)
					tForw2 = time.time() #end forward time, start loss time
					tLoss1 = tForw2
					tmpL, tmpLoss = self.computeLosses(y, yp, lossesList[r], relationList[r], weightRelations[r])
					errTot += tmpL.data.cpu()
					lossTot += tmpL
					for li, l in enumerate(tmpLoss):					
						errors[r][li] += l.data.cpu()
					tLoss2 = time.time() #end epoch time
					self.printBatchesLog(r, e, bi[r], errTotOld, errTot, len(datasetsList[r]), time.time()-tLoad1, tLoad2-tLoad1, tForw2-tForw1, tLoss2-tLoss1, start, batch_size, mute = MUTE)

				lossTot.backward()	
				optimizer.step()
				if FAST and c > 2:
					break
				c+=1
				tLoad1 = time.time()
				#if self.CUDA: #makes things slower
				#	t.cuda.empty_cache()
			end = time.time()	
			self.printLossLog(e, relationList, errors, start, end, errTotOld, errTot)
			if LOG:
				self.logger.writeTensorboardLog(e, errTot, errors, relationList, end-start)
			scheduler.step(float(errTot))
			self.saveModel(e)									
			e += 1	
		self.saveModel(True)
		if LOG:
			self.logger.shutdown()
	
	def predict(self, ERgraph, X, metaRelationName, relationName, sidex1=None, sidex2=None, batch_size=500, plotGraph=False):
		"""
		Function that performs the training of the wrapped pytorch model. It is analogous to scikit-learn .predict() method.



		Parameters
		----------
		ERgraph : ERgraph
		X : list
			List containing the 2D coordinates of the positions that should be predicted in the ERgraph.metaRelationName.relationName Relation.
		metaRelationName : str
			Name of the MetaRelation that contains the target relation
		relationName : str
			Name of the relation that you want to predict
		batch_size : int
			batch size during prediction

		Returns
		-------
		yp : list
			List containing the predictions for the target Relation

		"""
		targetMetaRel = ERgraph[metaRelationName]["metaRelation"]
		assert targetMetaRel != None, "ERROR: "+ metaRelationName+" is not a MetaRelation in the ER graph!"
		targetRelPos = targetMetaRel.getPos(relationName)
		assert targetRelPos != None, "ERROR: "+relationName+"is not a Relation in "+metaRelationName+" in the ERgraph!"
		print ("Rel in pos", targetRelPos)
		print ("self.DEVICE: ", self.dev)
		self.model.eval()
		print( "Training mode: ", self.model.training)
		if plotGraph:
			from pytorchUtils.torchgraphviz1 import make_dot, make_dot_from_trace
		
		print ("Predicting...")
		if sidex1 == None and sidex2 == None:
			dataset = D.PredictionDataset(X)
		else:
			dataset = D.PredictionDatasetSide(X, sidex1, sidex2)
		#predDataset = D.PredictMetaDataset([dataset], predRel.domain1, predRel.domain2, predRel.side1, predRel.side2)
		loader = DataLoader(dataset, batch_size=batch_size, shuffle=False, sampler=None, num_workers=0, collate_fn=Collate.predictMetaCollate)
		yp = []			
		first = True
		for sample in loader:
			#print len(sample)
			if len(sample) > 3:
				x1, x2, sx1, sx2 = sample
				x1 = t.tensor(x1, dtype=t.long, device=self.dev)
				x2 = t.tensor(x2, dtype=t.long, device=self.dev)
				if type(sx1) != type(None):
					sx1 = t.tensor(sx1, dtype=t.float, device=self.dev)
				if type(sx2) != type(None):
					sx2 = t.tensor(sx2, dtype=t.float, device=self.dev)
			else:
				x1, x2 = sample
				x1 = t.tensor(x1, dtype=t.long, device=self.dev)
				x2 = t.tensor(x2, dtype=t.long, device=self.dev)

			if len(sample) > 3:
				pred = self.model.forward(metaRelationName, x1, x2, NX.SafeVariable(sx1), NX.SafeVariable(sx2))
			else:
				pred = self.model.forward(metaRelationName, Variable(x1), Variable(x2))
			#pred = FC.getOutputs(pred, targetRelPos)
			#pred = t.sigmoid(pred)
			#print (pred.size())
			if first and plotGraph:
				first = False
				#print dict(self.model.named_parameters())
				#raw_input()
				make_dot(pred.mean(), params=dict(self.model.named_parameters()))		
			yp += pred.data[:,targetRelPos].cpu().squeeze().tolist()			
		return yp






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  PPIpredUtils.py
#  
#  Copyright 2017 Daniele Raimondi <eddiewrc@alice.it>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
import random, time, math, torch
import torch as t
import numpy as np
import NXTfusion.NXTfusion as NX
import NXTfusion.DataMatrix as DM
from multipledispatch import dispatch
from scipy.sparse import coo_matrix

def getOutputs(yp, c):
	"""
	:meta private:
	"""
	return yp[:,c:c+1]

def buildPytorchSide(data, domain, expectedLen = 20, sideDtype=np.float32):
	""" This function builds the data structure containing the side information 
	
		The data structure is a {} indicized with the domain numeric names.

		sideX = {domainName_numeric : numpy32_feats}
	:meta private:
	"""
	x = {}
	i = 0
	while i < len(domain):
		x[i] = []
		i+=1

	
	missing = 0
	for i in domain:
		#print i
		try:
			x[domain[i]].append(np.array(data[i], dtype=sideDtype))	
		except:
			x[domain[i]].append(np.zeros(expectedLen, dtype=sideDtype))	
			missing +=1
		#	print i, x[lookup[i]]
		#print len(x[lookup[i]][0]) #== 50
		#raw_input()
	if missing > 0:
		print( "WARNING: %d missing proteins in side info." % missing)
	return x


@dispatch(np.ndarray, NX.Entity, NX.Entity)
def buildPytorchFeats(data:np.ndarray, domain1:NX.Entity, domain2:NX.Entity, side1 = None, side2 = None) -> (list, list, list, list, list):
	"""
	This function is used to produce the input for the NNwrapper.predict() method, at prediction time. It produces the inputs necessary to predict the output for certain cells (or the entire matrix) for a given relation.

	This version of the method takes as prediction target a (dense) numpy matrix.

	Parameters
	----------
	data : np.ndarray
		Numpy matrix representing the target. This form of the method is more useful when the entire matrix needs to be predicted. The actual values in the matrix are provided as "labels" in output, but are ignored during prediction.
	domain1 : NX.Entity
		Entity representing the objects on the dimension 0 of the data matrix
	domain2 : NX.Entity
		Entity representing the objects on the dimension 1 of the data matrix
	
	Returns
	-------
	x : list of tuples [(i,j),(k,j),...]
		List containing the pairs of of objects belonging to domain1 and domain2 that needs to be predicted.
	y : list
		Values of the input data matrix corresponding to the pairs of object in x
	corresp: list of tuples
		List of tuples containing the corresponding names of the pairs of objects listed in x.
	"""
	assert data.shape[0] == len(domain1) and data.shape[1] == len(domain2)
	if side1 != None:
		assert len(side1) == data.shape[0]
	if side2 != None:
		assert len(side2) == data.shape[1]
	x = []
	y = []
	sx1 = []
	sx2 = []
	corresp = []
	
	i = 0
	while i < data.shape[0]:
		j = 0
		while j < data.shape[1]:
			#print i #(('Q96IY1', 'Q9UGM6'), 0)
			#raw_input()
			corresp.append((domain1[i], domain2[j]))
			x.append([i,j])
			if side1 != None:
				sx1.append(side1[i])
			if side2 != None:
				sx2.append(side2[j])
			y.append(data[i,j])
			j+=1
		i+=1
	assert len(x) == len(y)	
	if side1 == None:
		sx1 = None
	if side2 == None:
		sx2 == None
	if side1 == None and side2 == None:
		return x, y, corresp
	else:
		return x, y, sx1, sx2, corresp

@dispatch(dict, NX.Entity, NX.Entity)
def buildPytorchFeats(data:dict, domain1:NX.Entity, domain2:NX.Entity, side1 = None, side2 = None):
	"""
	This function is used to produce the input for the NNwrapper.predict() method, at prediction time. It produces the inputs necessary to predict the output for certain cells (or the entire matrix) for a given relation.

	This version of the method takes as prediction target a dict containing the pairs of ojects that need to be predicted. This is useful when only relatively few cells of the matrix need to be predicted (sparse prediction).

	Parameters
	----------
	data : dict {(obj[i],obj[j]):value1, (obj[i],obj[k]):value2, ...}
		Dict representing the target cells of the matrix that need to be predicted. The actual values in the matrix are provided as value associated to each pair of objects in the dict, but are ignored during prediction.
	domain1 : NX.Entity
		Entity representing the objects on the dimension 0 of the data matrix
	domain2 : NX.Entity
		Entity representing the objects on the dimension 1 of the data matrix
	
	Returns
	-------
	x : list of tuples [(i,j),(k,j),...]
		List containing the pairs of of objects belonging to domain1 and domain2 that needs to be predicted.
	y : list
		Values of the input data matrix corresponding to the pairs of object in x
	corresp: list of tuples
		List of tuples containing the corresponding names of the pairs of objects listed in x.
	"""
	x = []
	y = []
	sx1 = []
	sx2 = []
	corresp = []
	
	for i in data.items():
		#print i #(('Q96IY1', 'Q9UGM6'), 0)
		#raw_input()
		corresp.append(i[0])
		x.append([domain1[i[0][0]], domain2[i[0][1]]])
		if side1 != None:
			sx1.append(side1[domain1[i[0][0]]])
		if side2 != None:
			sx2.append(side2[domain2[i[0][1]]])
		y.append(i[1])
	assert len(x) == len(y)	
	if side1 == None:
		sx1 = None
	if side2 == None:
		sx2 == None
	if side1 == None and side2 == None:
		return x, y, corresp
	else:
		return x, y, sx1, sx2, corresp

@dispatch(DM.DataMatrix)
def buildPytorchFeats(datam:DM.DataMatrix, side1 = None, side2 = None):
	"""
	This function is used to produce the input for the NNwrapper.predict() method, at prediction time. It produces the inputs necessary to predict the output for certain cells (or the entire matrix) for a given relation.

	This version of the method takes as prediction target a DataMatrix object.

	Parameters
	----------
	datam : DataMatrix
		DataMatrix object containing the matrix representing the predidction target. The actual values in the observed cells in the DataMatrix are provided as "labels" y in output, but are ignored during prediction.
		
	Returns
	-------
	x : list of tuples [(i,j),(k,j),...]
		List containing the pairs of of objects belonging to domain1 and domain2 that needs to be predicted.
	y : list
		Values of the input data matrix corresponding to the pairs of object in x
	corresp: list of tuples
		List of tuples containing the corresponding names of the pairs of objects listed in x.
	"""
	data = datam.toHashTable()
	domain1 = datam.ent1
	domain2 = datam.ent2
	x = []
	y = []
	sx1 = []
	sx2 = []
	corresp = []
	
	for i in data.items():
		#print i #(('Q96IY1', 'Q9UGM6'), 0)
		#raw_input()
		corresp.append(i[0])
		x.append([domain1[i[0][0]], domain2[i[0][1]]])
		if side1 != None:
			sx1.append(side1[domain1[i[0][0]]])
		if side2 != None:
			sx2.append(side2[domain2[i[0][1]]])
		y.append(i[1])
	assert len(x) == len(y)	
	if side1 == None:
		sx1 = None
	if side2 == None:
		sx2 == None
	if side1 == None and side2 == None:
		return x, y, corresp
	else:
		return x, y, sx1, sx2, corresp

@dispatch(coo_matrix, NX.Entity, NX.Entity)
def buildPytorchFeats(data:coo_matrix, domain1:NX.Entity, domain2:NX.Entity, side1 = None, side2 = None):
	"""
	This function is used to produce the input for the NNwrapper.predict() method, at prediction time. It produces the inputs necessary to predict the output for certain cells (or the entire matrix) for a given relation.

	This version of the method takes as prediction target a dict containing the pairs of ojects that need to be predicted. This is useful when only relatively few cells of the matrix need to be predicted (sparse prediction).

	Parameters
	----------
	data : coo_matrix 
		Sparse matrix representing the target cells of the matrix that need to be predicted. The actual values in the matrix are provided as value associated to each pair of objects in the dict, but are ignored during prediction.
	domain1 : NX.Entity
		Entity representing the objects on the dimension 0 of the data matrix
	domain2 : NX.Entity
		Entity representing the objects on the dimension 1 of the data matrix
	
	Returns
	-------
	x : list of tuples [(i,j),(k,j),...]
		List containing the pairs of of objects belonging to domain1 and domain2 that needs to be predicted.
	y : list
		Values of the input data matrix corresponding to the pairs of object in x
	corresp: list of tuples
		List of tuples containing the corresponding names of the pairs of objects listed in x.
	"""
	x = []
	y = []
	sx1 = []
	sx2 = []
	corresp = []
	d = {}
	data = data.todok()
	for i in data.items():
		d[i[0]] = i[1]

	for i in data.items():
		#print i #(('Q96IY1', 'Q9UGM6'), 0)
		#raw_input()
		corresp.append(i[0])
		x.append([domain1[i[0][0]], domain2[i[0][1]]])
		if side1 != None:
			sx1.append(side1[domain1[i[0][0]]])
		if side2 != None:
			sx2.append(side2[domain2[i[0][1]]])
		y.append(i[1])
	assert len(x) == len(y)	
	if side1 == None:
		sx1 = None
	if side2 == None:
		sx2 == None
	if side1 == None and side2 == None:
		return x, y, corresp
	else:
		return x, y, sx1, sx2, corresp

def main():
	"""
	:meta private:
	"""	
	return 0

if __name__ == '__main__':
	main()




	
import torch as t
from torch.autograd import Variable
import os

class NXmodelProto(t.nn.Module):

	""" This class is the father of the pytorch modules used in the ER datafusion
	wrapper. It implements basic functions, leaving only the init and the forward empty"""

	def __init__(self):
		super(NXmodelProto, self).__init__()

	def getWeights(self):
		return self.state_dict()

	def getParameters(self):
		return list(self.parameters())

	def init_weights(self, m):
		if isinstance(m, t.nn.Conv1d) or isinstance(m, t.nn.Linear) or isinstance(m, t.nn.Bilinear):
			print ("Initializing weights...", m.__class__.__name__)
			t.nn.init.xavier_uniform(m.weight)
			m.bias.data.fill_(0.01)
		elif isinstance(m, t.nn.Embedding):
			print ("Initializing weights...", m.__class__.__name__)
			t.nn.init.xavier_uniform(m.weight)
		
	def getNumParams(self):
		p=[]
		for i in self.parameters():
			p+= list(i.data.cpu().numpy().flat)
		print ('Number of parameters=',len(p))	




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="NXTfusion", # Replace with your own username
    version="0.0.5",
    author="Daniele Raimondi",
    author_email="daniele.raimondi@kuleuven.be",
    description="Non linear data fusion over Entity Relation graphs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/eddiewrc/nxtfusion/src/master/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
	install_requires=['numpy', "scipy", "multipledispatch", "torch", "torchvision"],
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
)



# Code referenced from https://gist.github.com/gyglim/1f8dfb1b5c82627ae3efcfbbadb9f514
import numpy as np
import scipy.misc
import signal, os, sys
 
def signal_term_handler(signal, frame):
	os.system("killall tensorboard")
	print ('got SIGTERM, killing tensorboard.')
	sys.exit(0)
 
try:
    from StringIO import StringIO  # Python 2.7
except ImportError:
    from io import BytesIO         # Python 3.x

def to_np(x):
    return x.data.cpu().numpy()

def to_var(x):
    if torch.cuda.is_available():
        x = x.cuda()
    return Variable(x)   

class MetaLogger(object):

	""" This class incapsulates the tensorboard logging code,
	but uses the underlying Logger class to do all the real work."""

	def __init__(self, model, port = 6006):
		self.model = model
		import tensorflow as tf
		os.system("rm -rf ./logs")
		os.system("killall tensorboard")
		os.system("tensorboard --logdir='./logs' --port="+str(port)+"&")
		self.logger = Logger('./logs')			
		os.system("firefox http://127.0.0.1:"+str(port)+"  &")
		signal.signal(signal.SIGTERM, signal_term_handler)

	def writeTensorboardLog(self, step, errTot, lossScores, relationList, te):
		#step == epoch
		if self.logger == None:
			print (" WARNING: Logger not initializated, skipping")
			return
		info = {"errTot":errTot}
		for ri, r in enumerate(relationList):
			for i, l in enumerate(lossScores[ri]):
				info[r[i]["name"]] = l
		info["timePerEpoch"] = te

		for tag, value in info.items():
			self.logger.scalar_summary(tag, value, step+1)

		# (2) Log values and gradients of the parameters (histogram)
		for tag, value in self.model.named_parameters():
			#print
			#print tag#, value, value.grad					
			tag = tag.replace('.', '/')
			self.logger.histo_summary(tag, to_np(value), step+1)
			self.logger.histo_summary(tag+'/grad', to_np(value.grad), step+1)

		return info

	def shutdown(self):
		print ("Logger: killing tensorboard.")
		os.system("killall tensorboard")


class Logger(object):
    
	"""This class actually writes data on the tensorboard logger. It could be used
	directly or it can be used through the MetaLogger class"""

	def __init__(self, log_dir):
		"""Create a summary writer logging to log_dir."""
		self.writer = tf.summary.FileWriter(log_dir)
		tf.logging.set_verbosity(tf.logging.WARN)
		os.environ['TF_CPP_MIN_LOG_LEVEL']='5'

	def list_summary(self, tags, values, step):
		assert len(tags) == len(values)
		for i, t in enumerate(tags):
			self.scalar_summary(t, values[i], step)

	def scalar_summary(self, tag, value, step):
		"""Log a scalar variable."""
		summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
		self.writer.add_summary(summary, step)

	def image_summary(self, tag, images, step):
		"""Log a list of images."""

		img_summaries = []
		for i, img in enumerate(images):
			# Write the image to a string
			try:
				s = StringIO()
			except:
				s = BytesIO()
			scipy.misc.toimage(img).save(s, format="png")

			# Create an Image object
			img_sum = tf.Summary.Image(encoded_image_string=s.getvalue(),
									   height=img.shape[0],
									   width=img.shape[1])
			# Create a Summary value
			img_summaries.append(tf.Summary.Value(tag='%s/%d' % (tag, i), image=img_sum))

		# Create and write Summary
		summary = tf.Summary(value=img_summaries)
		self.writer.add_summary(summary, step)
		
	def histo_summary(self, tag, values, step, bins=1000):
		"""Log a histogram of the tensor of values."""

		# Create a histogram using numpy
		counts, bin_edges = np.histogram(values, bins=bins)

		# Fill the fields of the histogram proto
		hist = tf.HistogramProto()
		hist.min = float(np.min(values))
		hist.max = float(np.max(values))
		hist.num = int(np.prod(values.shape))
		hist.sum = float(np.sum(values))
		hist.sum_squares = float(np.sum(values**2))

		# Drop the start of the first bin
		bin_edges = bin_edges[1:]

		# Add bin edges and counts
		for edge in bin_edges:
			hist.bucket_limit.append(edge)
		for c in counts:
			hist.bucket.append(c)

		# Create and write Summary
		summary = tf.Summary(value=[tf.Summary.Value(tag=tag, histo=hist)])
		self.writer.add_summary(summary, step)
		self.writer.flush()
